package edu.kit.iti.formal.nonogram;

import javax.imageio.ImageIO;
import java.awt.image.BufferedImage;
import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.List;
import java.util.StringTokenizer;

/**
 * This class represents a Nonogram Riddle by its specification for row and column runs.
 *
 * @author Alexander Weigl
 * @version 1 (14.11.17)
 */
public class Nonogram {
    private String id;
    private List<List<Integer>> columnBlocks = new ArrayList<>();
    private List<List<Integer>> rowBlocks = new ArrayList<>();

    /**
     * Constructs a {@link Nonogram} form the given input string.
     * <p>
     * Refer to the excercise sheet for the encoding.
     *
     * @param input non null
     * @return a Nonogram
     * @see #toString()
     */
    public static Nonogram from(String input) {
        Nonogram m = new Nonogram();
        StringTokenizer tokenizer = new StringTokenizer(input, ":,;.!", true);
        m.id = tokenizer.nextToken().trim();
        List<List<Integer>> current = m.rowBlocks;
        ArrayList<Integer> block = null;
        while (tokenizer.hasMoreTokens()) {
            String token = tokenizer.nextToken();
            switch (token) {
                case ".":
                    current = m.columnBlocks;
                    //break; fall through
                case ";":
                case ":":
                    block = new ArrayList<>();
                    current.add(block);
                    break;
                case ",":
                    break;// do nothing
                case "!":
                    assert block != null;
                    return m;
                default:
                    int i = Integer.parseInt(token.trim());
                    assert block != null;
                    block.add(i);
                    break;
            }
        }
        throw new IllegalArgumentException("Input string not terminated with a bang »!«");
    }

    /**
     * Constructs a Nonogram from a solved puzzle.
     *
     * @param sol non null
     * @return non null nonogram
     */
    @SuppressWarnings("Duplicates")
    public static Nonogram from(boolean[][] sol) {
        Nonogram m = new Nonogram();
        for (int i = 0; i < sol.length; i++) {
            List<Integer> block = new ArrayList<>();
            int run = 0;
            for (int j = 0; j < sol[i].length; j++) {
                if (sol[i][j])
                    run++;
                else {
                    if (run != 0) {
                        block.add(run);
                    }
                    run = 0;
                }
            }
            if (run != 0)
                block.add(run);
            m.getRowBlocks().add(block);
        }

        for (int column = 0; column < sol[0].length; column++) {
            List<Integer> block = new ArrayList<>();
            int run = 0;
            for (int row = 0; row < sol.length; row++) {
                if (sol[row][column])
                    run++;
                else {
                    if (run != 0) {
                        block.add(run);
                    }
                    run = 0;
                }
            }
            if (run != 0)
                block.add(run);
            m.getColumnBlocks().add(block);
        }
        return m;
    }


    /**
     * Constructs a {@link Nonogram} from an image.
     *
     * The image should have 1-bit depth with black and white.
     *
     * @param image non null and existing
     * @return a nonnull reference
     * @throws IOException coming from reading the image
     */
    public static Nonogram from(File image) throws IOException {
        Nonogram a = from(ImageIO.read(image));
        a.id = image.getName();
        return a;
    }

    private static Nonogram from(BufferedImage read) {
        boolean[][] b = new boolean[read.getHeight()][read.getWidth()];
        int[] px = new int[4];

        for (int x = 0; x < read.getWidth(); x++) {
            for (int y = 0; y < read.getHeight(); y++) {
                read.getRaster().getPixel(x, y, px);
                b[y][x] = px[0] != 0;
            }
        }

        for (int i = 0; i < b.length; i++) {
            for (int j = 0; j < b[i].length; j++) {
                System.out.print(b[i][j] ? 'x' : '-');
            }
            System.out.println();
        }

        return from(b);
    }

    @Override
    public boolean equals(Object o) {
        if (this == o) return true;
        if (o == null || getClass() != o.getClass()) return false;

        Nonogram monogram = (Nonogram) o;

        if (columnBlocks != null ? !columnBlocks.equals(monogram.columnBlocks) : monogram.columnBlocks != null)
            return false;
        return rowBlocks != null ? rowBlocks.equals(monogram.rowBlocks) : monogram.rowBlocks == null;
    }

    @Override
    public int hashCode() {
        int result = columnBlocks != null ? columnBlocks.hashCode() : 0;
        result = 31 * result + (rowBlocks != null ? rowBlocks.hashCode() : 0);
        return result;
    }

    public int getWidth() {
        return columnBlocks.size();
    }

    public int getHeight() {
        return rowBlocks.size();
    }

    public List<List<Integer>> getColumnBlocks() {
        return columnBlocks;
    }

    public List<List<Integer>> getRowBlocks() {
        return rowBlocks;
    }

    public String getID() {
        return id;
    }

    @Override
    public String toString() {
        return getID() + ":" +
                sideToString(getRowBlocks())
                + "." + sideToString(getColumnBlocks()) + "!\n";
    }

    private String sideToString(List<List<Integer>> side) {
        return side.stream()
                .map(this::blockToString)
                .reduce((a, b) -> a + ";" + b)
                .orElse("");
    }

    private String blockToString(List<Integer> block) {
        return block.stream()
                .map(String::valueOf)
                .reduce((a, b) -> a + "," + b)
                .orElse("");
    }

}
