package edu.kit.iti.formal.nonogram;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileReader;
import java.io.IOException;
import java.lang.reflect.InvocationTargetException;
import java.util.ArrayList;
import java.util.List;

/**
 * @author Alexander Weigl
 * @vesion 2 (04.01.18)
 * @version 1 (14.11.17)
 */
public class Runner implements Runnable {
    public static char FIELD_WHITE = ' ';
    public static char FIELD_BLACK = 'x';
    public static boolean NO_CHECK = false;
    public static final int SECONDS = 1000;
    private List<Long> timings = new ArrayList<>();
    private volatile long startTime, stopTime;
    private int success;
    private int all;

    public static List<String> findRiddles() {
        final String riddlesFolder = System.getenv().
                getOrDefault("NONOGRAMS", "./nonograms.txt");
        final File file = new File(riddlesFolder);
        System.out.format("I am using the monograms in file '%s'%n", file);
        if (!file.exists()) {
            System.out.format("File %s does not exist. ABORT!%n", riddlesFolder);
            System.exit(1);
        }

        List<String> monograms = new ArrayList<>();
        try (BufferedReader br = new BufferedReader(new FileReader(file))) {
            String tmp = null;
            while (((tmp = br.readLine()) != null)) {
                if (!tmp.isEmpty())
                    monograms.add(tmp);
            }
        } catch (IOException e) {
            e.printStackTrace();
            System.exit(2);
        }
        return monograms;
    }


    @SuppressWarnings({"unchecked"})
    public static NonogramSolver solverInstance(Nonogram game)
            throws NoSuchMethodException, ClassNotFoundException,
            IllegalAccessException, InvocationTargetException,
            InstantiationException {
        Class<NonogramSolver> clazz = (Class<NonogramSolver>) Class.forName("MyNonogramSolver");
        NonogramSolver inst = clazz.getConstructor().newInstance();
        inst.init(game);
        return inst;
    }

    public static void main(String[] args) {
        Runner runner = new Runner();
        NO_CHECK = Boolean.valueOf(
                System.getenv().getOrDefault("NO_CHECK", "false"));
        int time = Integer.valueOf(System.getenv().getOrDefault("TIMEOUT", "100"));

        System.out.println("Solution checking is " + (NO_CHECK ? "disabled" : "enabled") + ".");
        System.out.println("Timeout is  " + time + " seconds.");

        Thread t = new Thread(runner);
        t.start();
        try {
            t.join(time * SECONDS);
        } catch (InterruptedException e) {
            //
        } finally {
            runner.statistics();
            System.exit(0);
        }
    }

    static boolean printAndCheckSolution(Nonogram monogram, boolean[][] sol) {
        System.out.format("> SOLUTION %s%n", monogram.getID());
        for (int i = 0; i < sol.length; i++) {
            for (int j = 0; j < sol[i].length; j++) {
                System.out.print(sol[i][j] ? FIELD_BLACK : FIELD_WHITE);
            }
            System.out.println();
        }
        System.out.format("<END SOLUTION %s%n", monogram.getID());

        Nonogram m = Nonogram.from(sol);
        return monogram.equals(m);
    }

    static void statistics(long success, long all, long startTime, long stopTime, List<Long> timings) {
        System.out.format("%%%% %s: %d of %d (%2.1f%%) successful!%n",
                new File(".").getAbsolutePath(),
                success, all, 100.0 * success / all);

        System.out.printf("%%%% %s: %10d ms%n",
                new File(".").getAbsolutePath(),
                stopTime - startTime);

        if (timings.isEmpty())
            return;

        long max = timings.stream().max(Long::compare).get();
        long min = timings.stream().min(Long::compare).get();

        long sum = timings.stream().reduce((a, b) -> a + b).get();

        System.out.printf("%% Sum time: %10d ms%n", sum);
        System.out.printf("%% Max time: %10d ms%n", max);
        System.out.printf("%% Min time: %10d ms%n", min);
        System.out.printf("%% Avg time: %10.0f ms%n", sum / (float) timings.size());
        System.out.println("%% end");
    }

    public boolean checkSolver(String puzzle) throws ClassNotFoundException, NoSuchMethodException, InvocationTargetException, InstantiationException, IllegalAccessException {
        Nonogram monogram = Nonogram.from(puzzle);

        System.out.printf("== %s ======================================================%n", monogram.getID());
        NonogramSolver solver = solverInstance(monogram);

        long start = System.currentTimeMillis();
        boolean ans = solver.solve();
        timings.add(System.currentTimeMillis() - start);

        boolean satisfiable = !monogram.getID().startsWith("u");

        System.out.println("Riddle is " + (!satisfiable ? "not " : "") + "satisfiable");
        System.out.println("Your answer is " + (ans ? "sat" : "unsat"));

        if (ans && !NO_CHECK) {
            if (! printAndCheckSolution(monogram, solver.getSolution())) {
                System.out.println("There is an error in your solution!");
                return false;
            }
        }

        if (ans == satisfiable)
            System.out.println("SOLVED!");
        else
            System.out.println("NOT SOLVED :(");

        return ans == satisfiable;
    }

    public void run() {
        try {
            startTime = System.currentTimeMillis();
            success = 0;
            all = 0;
            for (String s : findRiddles()) {
                all++;
                success += checkSolver(s) ? 1 : 0;
            }
        } catch (ClassNotFoundException e) {
            System.err.println("Could not from the solution by class name.");
            System.err.println("There should be class 'MyNonogramSolver' in the default package on the classpath.");
            e.printStackTrace();
            System.exit(3);
        } catch (NoSuchMethodException | InvocationTargetException | InstantiationException | IllegalAccessException e) {
            System.err.println("Constructor not found");
            //System.err.println("There should be constructor 'public MyKuromasuSolver(Kuromasu)' in your solution class.");
            System.err.println(e.getMessage());
            e.printStackTrace();
            System.exit(4);
        } finally {
            stopTime = System.currentTimeMillis();
        }
    }

    public void statistics() {
        statistics(success, all, startTime, stopTime, timings);
    }
}
